#ifndef COMMON_H
#define COMMON_H

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <pthread.h>
#include <errno.h> 
#include <string.h>
#include "lzf.h"
#include "adoc.h"
#include <sys/uio.h>
#include "config.h"
#include "adoc_sem.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <inttypes.h>
#include "zlib.h"
#include "timings.h"

/*8 bits type*/
typedef int8_t byte;



/******  ADOC Internal Constant **************/


/* BS = Begin Session*/
#define ADOC_TYPE_BS 1
#define ADOC_TYPE_COMP 2
#define ADOC_TYPE_UNCOMP 3
#define ADOC_TYPE_LAST_PACKET 4
#define ADOC_TYPE_EOF 5
#define ADOC_TYPE_SESSION_SIZE 6

/* The 3 following #define MUST be negative and coded by a 8 bits type*/
#define ADOC_RAW -1
#define ADOC_CMP -2
#define ADOC_TEST -3


/*All these define must be coded by a 4 bytes type*/
#define ADOC_PACKET_SIZE (8*1024)
#define ADOC_BUFFER_SIZE (256*1024)
#define LZF_BUFFER_SIZE (32*1024)
#define TCP_TIMING_SIZE (256*1024)
#define TCP_BUFFER_SIZE (32*1024)
/*TCP_PIECE_SIZE must divide TCP_TIMING_SIZE*/ 
#define TCP_PIECE_SIZE (TCP_TIMING_SIZE/16)
#define ADOC_MAX_NB_PACKETS 30


#define CHECK_ERR(err, msg) { \
    if (err != Z_OK) { \
        fprintf(stderr, "%s error: %d\n", msg, err); \
        ADOC_global->error=ADOC_COMPRESSION_ERROR; \
        return -1; \
    } \
}

#define CHECK_RETVAL(err, msg) { \
    if (err != 0) { \
        fprintf(stderr, "%s error: %d (line: %d, file :%s)\n", msg, err,__LINE__,__FILE__); \
        ADOC_global->error=ADOC_RETVAL_ERROR; \
        pthread_exit(0); \
    } \
}



typedef struct _stat{
    int level;
    int fifo_size;
    int taille;
}adoc_stat;

/* We use a circular window for storing packets
 */
typedef struct {
    byte type;
    byte level;
    size_t data_size;
    size_t uncomp_size;
    byte *data;
}header;


#define ADOC_MIN(a,b) ((a) <= (b) ? (a) : (b))
#define ADOC_MAX(a,b) ((a) >= (b) ? (a) : (b))
#define READ(fd,buf,count) my_read(fd,buf,count,__LINE__,__FILE__)
#define WRITE(fd,buf,count) my_write(fd,buf,count,__LINE__,__FILE__)



ssize_t my_read(int, void *, size_t ,int ,char *);
ssize_t my_write(int, void *, size_t ,int ,char *);




#endif /*COMMON_H*/



