#include "common.h"
#include "reception.h"

/* the buffer list is the only global variable, 
in order the program to be thread safe, this variable 
is accessible only with a mutex*/
static list *ADOC_buffer_list=NULL;
static pthread_mutex_t ADOC_buffer_list_mutex=PTHREAD_MUTEX_INITIALIZER;

int store_buffer(int socket,byte * buffer,size_t size,ADOC_session *ADOC_global){
    list *cell;

    
    cell=(list*)malloc(sizeof(list));
    if(cell==NULL){
        errno=ENOMEM;
        ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
        return -1;
    }
    if(ADOC_global->store_size!=0){
        byte *new_buffer,*store_buffer=ADOC_global->store_buffer;
        size_t store_size=ADOC_global->store_size;

        new_buffer=(byte *)malloc(sizeof(byte)*(size+store_size));
        if(new_buffer==NULL){
            ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
            return -1;
        }		
 
        memcpy(new_buffer,store_buffer,store_size);
        memcpy(new_buffer+store_size,buffer,size);
        free(store_buffer);
        free(buffer);
        buffer=new_buffer;
        size+=store_size;
    }

    cell->socket=socket;
    cell->buffer=buffer;
    cell->size=size;

    pthread_mutex_lock(&ADOC_buffer_list_mutex);
    cell->next=ADOC_buffer_list;
    ADOC_buffer_list=cell;
    pthread_mutex_unlock(&ADOC_buffer_list_mutex);
            
    return 0;
}

int store_buffer2(int socket,byte * buffer,size_t size,byte *s_buffer,size_t store_size){
    ADOC_session ADOC_global;
    ADOC_global.store_buffer=s_buffer;
    ADOC_global.store_size=store_size;
    return store_buffer(socket,buffer,size,&ADOC_global);
}

ssize_t retreive_buffer(int socket,byte * buffer, size_t size){
    list *cell,*prev_cell=NULL;
    byte * new_buffer;
    ssize_t new_size;

    pthread_mutex_lock(&ADOC_buffer_list_mutex);
    if(ADOC_buffer_list==NULL){
        pthread_mutex_unlock(&ADOC_buffer_list_mutex);
        return 0;
    }
    cell=ADOC_buffer_list;
    do{
        if(cell->socket==socket){ /*find THE cell that has the good socket ID*/
            if(size<cell->size){/*retreive the buffer partially*/
                memcpy(buffer,cell->buffer,size);
                new_size=cell->size-size;
                new_buffer=(byte *)malloc(new_size);
                if(new_buffer==NULL){
                    errno=ENOMEM;
                    return ADOC_ERROR_WITH_ERRNO_SET;;
                }
                memcpy(new_buffer,cell->buffer+size,new_size);
                free(cell->buffer);
                cell->buffer=new_buffer;
                cell->size=new_size;
                pthread_mutex_unlock(&ADOC_buffer_list_mutex);
                return size;
            }else{ /*retreive the whole buffer and free memory */
                size=cell->size;
                memcpy(buffer,cell->buffer,size);
                free(cell->buffer);
                if(prev_cell!=NULL)
                    prev_cell->next=cell->next;
                else{
                    ADOC_buffer_list=cell->next;
                }
                pthread_mutex_unlock(&ADOC_buffer_list_mutex);
                free(cell);
                return size;
            }
        }
prev_cell=cell;
cell=cell->next;
    }while(cell!=NULL);
pthread_mutex_unlock(&ADOC_buffer_list_mutex);
return 0;
}

void free_buffer(int socket){
    list *cell,*prev_cell=NULL;
    pthread_mutex_lock(&ADOC_buffer_list_mutex);
    if(ADOC_buffer_list==NULL){
        pthread_mutex_unlock(&ADOC_buffer_list_mutex);
        return;
    }
    cell=ADOC_buffer_list;

    do{
        if(cell->socket==socket){
            free(cell->buffer);
            if(prev_cell!=NULL)
                prev_cell->next=cell->next;
            else
                ADOC_buffer_list=cell->next;
            free(cell);
            pthread_mutex_unlock(&ADOC_buffer_list_mutex);
            return;
        }
        prev_cell=cell;
        cell=cell->next;
    }while(cell!=NULL);
    pthread_mutex_unlock(&ADOC_buffer_list_mutex);

}



