#include <string.h>
#include "adoc.h"
#include "timings.h"
#include "config.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h> 
#include <netinet/in.h>
#ifndef NO_SOCKET_TRICK
#include <netinet/ip.h>
#endif
#include <netinet/tcp.h>
#include <string.h>
#include <netdb.h>
#ifdef NO_SOCKET_TRICK
#include <strings.h>
#endif

FILE *adoc_fopen(char *nom,char *mode){
    FILE *res;
    res=fopen(nom,mode);
    if(res==NULL){
	fprintf(stderr,"Unable to open %s with mode %s!\n",nom,mode);
	exit(-1);
    }
    return res;
}


int connect_server(int port,char *host){
  int sock;
  struct sockaddr_in nom;
  struct hostent *hp, *gethostbyname();

#ifndef NO_SOCKET_TRICK
  int tos;
#endif


  if((sock=socket(AF_INET,SOCK_STREAM,0))==-1){
    perror("Creating socket failed.\n");
    return -2;
  }
#ifndef NO_SOCKET_TRICK
  tos = IPTOS_THROUGHPUT;
 if(setsockopt(sock, IPPROTO_IP, IP_TOS, (char *)&tos,sizeof(tos))<0)
     printf("TOS not supported!\n");;
#endif
  /*recherche de l'adresse internet du serveur*/

  if((hp=gethostbyname(host))==NULL){
    fprintf(stderr,"Host : %s unknown.\n",host);
    return -3;
  }
  

  bcopy(hp->h_addr,
	&nom.sin_addr,
	hp->h_length);
  nom.sin_family=AF_INET;
  nom.sin_port=htons(port);
  

  /* Asking for connection */

  if((connect(sock,(struct sockaddr *)&nom,sizeof(nom)))==-1){
    perror("Cannot connect to serveur 1\n");
    return -4;
  }

  return sock;
}

int main(int argc,char *argv[]){
    int port,socket;
    char *host;
    FILE *pf;
    CLOCK_T begin,end;
    double time;
    int min_level=0,max_level=10;
    size_t file_size;
    size_t sent_size;
    if(argc<4){
	fprintf(stderr,"usage : %s host port file_name [min_level [max_level]]\n",argv[0]);
	exit(-1);
    }
  
    host=strdup(argv[1]);
    port=atoi(argv[2]);
    pf=adoc_fopen(argv[3],"r");
    if(argc>4)
	min_level=atoi(argv[4]);
    if(argc>5)
	max_level=atoi(argv[5]);

    socket=connect_server(port,host);
    if(socket<0){
	fprintf(stderr,"Connection to server failed... Aborting.\n");
	exit(-1);
    }
    
    printf("file: %s\n",argv[3]);
    printf("Host: %s:%d\n",host,port);
    printf("Min compression level=%d\n",min_level);
    printf("Max compression level=%d\n",max_level);

    CLOCK(begin);
    file_size=adoc_send_file_levels(socket,pf,&sent_size,min_level,max_level);
    CLOCK(end);  
    time=CLOCK_DIFF(end,begin);
    fclose(pf);
    printf("Sending time = %.2f seconds.\n",time);
    printf("Data size = %lu.\n",(unsigned long int)file_size);
    printf("Sent size = %lu.\n",(unsigned long int)sent_size);
    printf("Network throughput = %.2f Mbit/s.\n",sent_size*8/(time*1000000));
    printf("Application throughput = %.2f Mbit/s.\n",file_size*8/(time*1000000));
    printf("Average compression ratio = %.2f\n",(double)file_size/(double)sent_size);
    adoc_close(socket);

    return 0;
}

