#include "common.h"
#include "emission.h"


#ifdef DEBUG
void display_stat(adoc_stat *history,int n){
    int i=0;
    int j;
    printf("\n\n--------- Record ----------\n");
    while(i<n){
        if((history[i].level==0)){
            j=1;
            i++;
            while((history[i].level==0)&&(i<n)){
                j++;i++;
            }
            printf("%d packet level 0\n",j);
        }else{
            printf("level=%d, data_size=%d, fifo_size=%d\n",history[i].level,
                   history[i].taille, history[i].fifo_size);
            i++;
        }
    }
    printf("\n");
}
#endif

#ifdef DEBUG
void display_timings(ADOC_session *ADOC_global){
    int i;

    fprintf(stderr,"-----------Speed table-------------\n");

    for(i=0;i<ADOC_global->max_level+1;i++)
        fprintf(stderr,"min_speed[%d]=%.3f\n",i,ADOC_global->speed_min[i]);
    fprintf(stderr,"\n");
}

#endif

void update_send_window(byte type,byte level, size_t data_size,
                        size_t uncomp_size,ADOC_session *ADOC_global){
    int rval;

    /* Implementation ensure that if you are here, there is at least
    enough space in the window to store one packet*/

#ifdef DUMP_PACKET
    printf("packet[%d] : type: %d, level: %d, size: %ld\n",ADOC_global->header_w,type,level,data_size);
#endif
    ADOC_global->header_window[ADOC_global->header_w].type=type;
    ADOC_global->header_window[ADOC_global->header_w].level=level;
    ADOC_global->header_window[ADOC_global->header_w].data_size=data_size;
    ADOC_global->header_window[ADOC_global->header_w].data=ADOC_global->data_w_pointer;
    ADOC_global->header_window[ADOC_global->header_w].uncomp_size=(uint32_t)uncomp_size;
    /* Tell the other thread that one packet has been added to the window*/
    rval=adoc_sem_post(&(ADOC_global->r_sem));
    CHECK_RETVAL(rval,"sem_post");

    /*wait if the window is full*/
    rval=adoc_sem_wait(&(ADOC_global->w_sem));
    CHECK_RETVAL(rval,"sem_wait");
    /*Incrementing the size*/
    pthread_mutex_lock (&(ADOC_global->fifo_mutex));
    ADOC_global->fifo_size++;
    pthread_mutex_unlock (&(ADOC_global->fifo_mutex));

    /*Updating the writing pointer of the data window cyclically*/
    ADOC_global->data_w_pointer+=ADOC_PACKET_SIZE;
    /* If dealing with an uncompressed packet or the last packet of a compressed
        buffer or etc... */
    if((type!=ADOC_TYPE_COMP) && (ADOC_global->data_w_pointer>=ADOC_global->data_window+ADOC_MAX_SEND_PACKETS*ADOC_PACKET_SIZE))
        ADOC_global->data_w_pointer=ADOC_global->data_window;

    /*Updating the writing pointer of the packet window cyclically */
    ADOC_global->header_w++;
    if(ADOC_global->header_w>=ADOC_MAX_SEND_PACKETS)
        ADOC_global->header_w=0;

}




/* level must be between 2 and 10*/

/*return 0 in case of success
return -1 in case of failure
return 10 if compression is not usefull*/

int zlib_window_compression(Bytef *uncomp_buffer,size_t size_in,byte level,ADOC_session *ADOC_global){
    z_stream  c_stream;
    int err;
    uLongf comp_size,tot_comp_size=0;
    double comp_ratio;


    c_stream.zalloc = (alloc_func)0;
    c_stream.zfree = (free_func)0;
    c_stream.opaque = (voidpf)0;
    /*zlib level compression between 1 and 9*/
    err = deflateInit(&c_stream, level-1);
    CHECK_ERR(err, "deflateInit");
    c_stream.next_in=uncomp_buffer;
    /* After ADOC_data_w_pointer, there is always enough spoace to store the largest
        compressed buffer*/
    c_stream.next_out=(Bytef*)ADOC_global->data_w_pointer;
    c_stream.avail_in=(uInt)size_in;
    /*printf("\nuncomp_size=%d\n",size_in);*/
    /* Compressing the buffer */
    do{
        c_stream.avail_out = (uInt)ADOC_PACKET_SIZE;
        err = deflate(&c_stream,Z_FINISH);
        comp_size = ADOC_PACKET_SIZE-c_stream.avail_out;
        if(err==Z_STREAM_END)
            update_send_window(ADOC_TYPE_LAST_PACKET,level,comp_size,size_in,ADOC_global);
        else{
            if(c_stream.avail_out!=0){
                fprintf(stderr,"AdOC : Generated packet too small line %d!\n",__LINE__);
                ADOC_global->error=ADOC_COMPRESSION_ERROR;
                return -1;
            } else
                update_send_window(ADOC_TYPE_COMP,level,comp_size,0,ADOC_global);
        }
        tot_comp_size+=comp_size;
    }while(err==Z_OK);
    /*printf("comp_size=%ld\n",dec);*/
    if(err!=Z_STREAM_END)
        CHECK_ERR(err, "deflate");
    fflush(stdout);
    deflateEnd(&c_stream);
    comp_ratio=((double)size_in)/((double)tot_comp_size);
    /*if compression ratio is lower than 1.2 do not compress anymore*/
    if(comp_ratio<1.2){
      return 10;
    }
    return 0;
}


/*compression for level 1*/

/*return 0 in case of success
return -1 in case of failure
return 10 if compression is not usefull*/
int lzf_window_compression(byte *uncomp_buffer,size_t size_in,ADOC_session *ADOC_global){

    size_t comp_size,i,in_len;
    int chuck_size;
    while(size_in>0){
        in_len=ADOC_MIN(size_in,LZF_BUFFER_SIZE);
        comp_size=lzf_compress(uncomp_buffer,in_len,ADOC_global->data_w_pointer,2*in_len);
        if(comp_size==0){
            ADOC_global->error=ADOC_COMPRESSION_ERROR;
            return -1;
        }
        /* store intermediate packets */
        for(i=0;i<comp_size/ADOC_PACKET_SIZE;i++){
#ifdef DUMP_PACKET
            printf("++data: %d (%d) %d (%d) \n",ADOC_global->data_w_pointer[0],(int)(&(ADOC_global->data_w_pointer[0])),
                   ADOC_global->data_w_pointer[1],(int)(&(ADOC_global->data_w_pointer[1])));
#endif
            update_send_window(ADOC_TYPE_COMP,1,ADOC_PACKET_SIZE,0,ADOC_global);
        }
#ifdef DUMP_PACKET
        printf("++data: %d (%d) %d (%d) \n",ADOC_global->data_w_pointer[0],(int)(&(ADOC_global->data_w_pointer[0])),
               ADOC_global->data_w_pointer[1],(int)(&(ADOC_global->data_w_pointer[1])));
#endif
        update_send_window(ADOC_TYPE_LAST_PACKET,1,comp_size%ADOC_PACKET_SIZE,in_len,ADOC_global);


        size_in-=in_len;
        uncomp_buffer+=in_len;

        if(comp_size>in_len){
            while(size_in>0){
                chuck_size=ADOC_MIN(ADOC_PACKET_SIZE,size_in);
                memcpy(ADOC_global->data_w_pointer,uncomp_buffer,chuck_size);
                update_send_window(ADOC_TYPE_UNCOMP,0,chuck_size,chuck_size,ADOC_global);
                size_in-=chuck_size;
                uncomp_buffer+=chuck_size;
            }
            return 10;
        }
    }
    return 0;
}



/* Compress data and store them into the data window*/

/*return 0 in case of success
return -1 in case of failure
return 1 if compression is not usefull*/

int window_compression(byte *uncomp_buffer,size_t size_in,byte level,ADOC_session *ADOC_global){
    if(level==1)
        return lzf_window_compression(uncomp_buffer,size_in,ADOC_global);
    else{

        return zlib_window_compression((Bytef *)uncomp_buffer,size_in,level,ADOC_global);

    }
}


byte  update_level(int level,int prev_fifo_size,ADOC_session *ADOC_global){

    int diff,nb_cells;
    byte best_level;
    double best_speed;
    int i;

   if((ADOC_global->speed_min[0]==-1)&&(level!=0)&&(ADOC_global->min_level==0)){
        return level;
    }
    diff=ADOC_global->fifo_size-prev_fifo_size;
    nb_cells=ADOC_global->fifo_size;
    /* Update compression level*/
    if(nb_cells==0)
        return ADOC_global->min_level;
    if(nb_cells<10){
        if(diff<=0)
            level=level/2;
    }else if(nb_cells<20){
        if((diff>1)||((diff>0)&&(level<=1))){ /*dont take control packets into consideration*/
            level++;
        } else if (diff<0){
            if(level>0)
                level--;
        }
    }else if(nb_cells<=30){
        if((diff>1)||((diff>0)&&(level<=1))){
            if(level==0)
                level+=1;
            else
                level+=2;
        }else if(diff<0){
            level--;
        }
    }else if((diff>1)||((diff>0)&&(level<=1))){
        if(level==0)
            level+=1;
        else
            level+=2;
    }


if (level>ADOC_global->max_level)
level=ADOC_global->max_level;
else if (level<ADOC_global->min_level)
level=ADOC_global->min_level;

    if(!ADOC_global->reset_speed_min[level]){
       best_level=ADOC_global->min_level;
       best_speed=ADOC_global->speed_min[best_level];
       for(i=ADOC_global->min_level+1;i<=level;i++){
           if(best_speed<ADOC_global->speed_min[i]){
               best_speed=ADOC_global->speed_min[i];
               best_level=i;
           }
       }
    level=best_level;
    }	
           
        /*while(level>ADOC_global->min_level){
            sub_level=level-1;
            while(sub_level>ADOC_global->min_level){
                if(ADOC_global->speed_min[sub_level]==-1)
                    sub_level--;
                else
                    break;
            }

            if(ADOC_global->speed_min[sub_level]==-1)
                break;

            if(ADOC_global->speed_min[sub_level]>ADOC_global->speed_min[level]){
                level=sub_level;
            }else
                break;
        }*/


return (byte)level;
}


void *data_compression_thread (void *arg){


    byte level;
    int chuck_size,prev_fifo_size;
    byte *data_buffer;
    size_t data_size;
    int rval,res=0;

    ADOC_session *ADOC_global=(ADOC_session*)arg;

#ifdef DEBUG
    int h=0;
    adoc_stat *history;


    history=(adoc_stat*)malloc(100000*sizeof(adoc_stat));
    if(history==NULL){
        errno=ENOMEM;
        ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
        pthread_exit(0);
    }
#endif

    data_buffer=ADOC_global->data_buffer;
    data_size=ADOC_global->data_size;
    /*Updating the max_compression level in order both party to have the same value*/

    update_send_window(ADOC_TYPE_BS,0,ADOC_BUFFER_SIZE,0,ADOC_global);
    update_send_window(ADOC_TYPE_SESSION_SIZE,0,data_size,0,ADOC_global);

    ADOC_global->uncomp_size= 0;
    level=ADOC_global->min_level;
    do{

        prev_fifo_size=ADOC_global->fifo_size;
#ifdef DEBUG
        history[h].fifo_size=prev_fifo_size;
#endif

        if(level){
            if((chuck_size=ADOC_MIN(ADOC_BUFFER_SIZE,data_size))>=0){
                if((res=window_compression(data_buffer,chuck_size,level,ADOC_global))<0)
                    pthread_exit(0);
                data_size-=chuck_size;
                ADOC_global->uncomp_size+=chuck_size;
                data_buffer+=chuck_size;

            }else {
                data_size=0;
            }


        } else {
            if((chuck_size=ADOC_MIN(ADOC_PACKET_SIZE,data_size))>=0){
                memcpy(ADOC_global->data_w_pointer,data_buffer,chuck_size);
                update_send_window(ADOC_TYPE_UNCOMP,0,chuck_size,chuck_size,ADOC_global);
                data_size-=chuck_size;
                ADOC_global->uncomp_size+=chuck_size;
                data_buffer+=chuck_size;
            }else {
                data_size=0;
            }


        }
#ifdef DEBUG
        history[h].level=level;
        history[h].taille=chuck_size;
        h++;
#endif
        /*if window_compression return 1 compression is not usefull because data cannot compress*/
        if(res>0){
            level=ADOC_global->min_level;
            res--;
        }else
            level=update_level(level,prev_fifo_size,ADOC_global);

        /*Check whenever an error has been generated by
            the emission thread*/
        if(ADOC_global->error!=0){
            pthread_exit(0);
        }
    }while(data_size>0);
    update_send_window(ADOC_TYPE_EOF,0,0,0,ADOC_global);
    rval=adoc_sem_wait(&(ADOC_global->history_sem));
    CHECK_RETVAL(rval,"sem_wait");
#ifdef DEBUG
    display_stat(history,h);
    free(history);
#endif
    pthread_exit(0);
    /* Make comiler happy!*/
    return NULL;
}


void *file_compression_thread (void *arg){
    byte level;
    int chuck_size,prev_fifo_size;
    byte *uncomp_buffer;
    FILE *pf;
    int rval;
    int res=0;

    ADOC_session *ADOC_global=(ADOC_session*)arg;

#ifdef DEBUG
    int h=0;
    adoc_stat *history;


    history=(adoc_stat*)malloc(100000*sizeof(adoc_stat));
    if(history==NULL){
        errno=ENOMEM;
        ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
        pthread_exit(0);
    }

#endif



    pf=ADOC_global->file;

    /* Set ADOC_global variable size */
    /*fseek(pf,0,SEEK_END);
    ADOC_global->file_size=ftell(pf);
    fseek(pf,start,SEEK_SET);
    printf("File size : %d (%d,%d)\n", ADOC_global->file_size,start,ftell(pf));*/




    uncomp_buffer=(byte *)malloc(ADOC_BUFFER_SIZE);
    if(uncomp_buffer==NULL){
        errno=ENOMEM;
        ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
#ifdef DEBUG
        free(history);
#endif
        pthread_exit(0);
    }

    update_send_window(ADOC_TYPE_BS,0,ADOC_BUFFER_SIZE,0,ADOC_global);

    level=ADOC_global->min_level;
    do{

        prev_fifo_size=ADOC_global->fifo_size;
#ifdef DEBUG
        /*printf("size =%d\n",prev_fifo_size);*/
        history[h].fifo_size=prev_fifo_size;
#endif

        if(level){
            if((chuck_size=read(fileno(pf),uncomp_buffer,ADOC_BUFFER_SIZE))>0){
                if((res=window_compression(uncomp_buffer,chuck_size,level,ADOC_global))<0){
                    free(uncomp_buffer);
#ifdef DEBUG
                    free(history);
#endif
                    pthread_exit(0);
                }
            }else {
                if(chuck_size<0){
                    fprintf(stderr,"AdOC : Error while reading file : %s\n",strerror(errno));
                    ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
                    free(uncomp_buffer);
#ifdef DEBUG
                    free(history);
#endif
                    pthread_exit(0);
                }
            }


        } else {
            if((chuck_size=read(fileno(pf),ADOC_global->data_w_pointer,ADOC_PACKET_SIZE))>0){
                /*WRITE(ADOC_global->socket,ADOC_data_w_pointer,ADOC_PACKET_SIZE);*/
                update_send_window(ADOC_TYPE_UNCOMP,0,chuck_size,chuck_size,ADOC_global);
            }else {
                chuck_size=read(fileno(pf),ADOC_global->data_w_pointer,ADOC_PACKET_SIZE);
                if(chuck_size<0){
                    fprintf(stderr,"AdOC : Error while reading file : %s\n",strerror(errno));
                    ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
                    free(uncomp_buffer);
#ifdef DEBUG
                    free(history);
#endif
                    pthread_exit(0);
                }
            }

        }
#ifdef DEBUG
        history[h].level=level;
        history[h].taille=chuck_size;
        h++;
#endif
        /*if window_compression return 1 compression is not usefull because data cannot compress*/
        if(res>0){
            level=ADOC_global->min_level;
            res--;            
        }else
            level=update_level(level,prev_fifo_size,ADOC_global);

        /*Check whenever an error has been generated by
            the emission thread*/
        if(ADOC_global->error!=0){
            free(uncomp_buffer);
#ifdef DEBUG
            free(history);
#endif

            pthread_exit(0);
        }

    }while(chuck_size);
    update_send_window(ADOC_TYPE_EOF,0,0,0,ADOC_global);
    free(uncomp_buffer);
    rval=adoc_sem_wait(&(ADOC_global->history_sem));
    CHECK_RETVAL(rval,"sem_wait");


#ifdef DEBUG
    display_stat(history,h);
    free(history);
    display_timings(ADOC_global);
#endif
    

    pthread_exit(0);
    /* Make comiler happy!*/
    return NULL;
}




