#include "emission.h"
#include "timings.h"
#undef DUMP_PACKET


void store_timing(ADOC_session *ADOC_global,CLOCK_T start,CLOCK_T end,size_t buffer_size,int level,int nb_packets,size_t sent_size){
    double speed,time;
    int i;

    if(CLOCK_DIFF(end,ADOC_global->time_stamp)>=2.0){
        for(i=0;i<ADOC_global->max_level+1;i++)
            ADOC_global->reset_speed_min[i]=1;
        ADOC_global->time_stamp=end;
    }


        time=CLOCK_DIFF(end,start);
        speed=((double)buffer_size*8)/(time*1000000); /*in Mbit/s*/


        if((speed>ADOC_global->speed_min[level])&&(ADOC_global->reset_speed_min[level])){
            ADOC_global->speed_min[level]=speed;


        } else if((speed<ADOC_global->speed_min[level])){
            
            ADOC_global->speed_min[level]=speed;
                
        }
}


/*This thread has in charge of reading packets from the
circular window and send them to the network*/

void *emission_thread(void *arg){
    size_t data_size,prev_data_size=0,sent_size=0;
    uint32_t data_size_t;
    uint8_t type_t,level_t;
    int prev_type=-1,prev_level=0,start_clock=0,level;
    CLOCK_T start,end;
    int nb_packets=0;
    size_t buffer_size=0;
#ifdef VERBOSE
    size_t sent=0;
#endif
    int socket;
    struct iovec iov[4];
    int type,res,rval;
#ifdef DUMP_PACKET
    int i;
#endif

    ADOC_session *ADOC_global=(ADOC_session*)arg;
    socket=ADOC_global->socket;
#ifdef VERBOSE
    printf("\n 0.00%% sent");
    fflush(stdout);
#endif
    iov[3].iov_len=0;
    iov[3].iov_base=NULL;
    do{
        /* wait for having at least one packet to read*/
        rval=adoc_sem_wait(&(ADOC_global->r_sem));
        CHECK_RETVAL(rval,"sem_wait");
        
        data_size=ADOC_global->header_window[ADOC_global->header_r].data_size;
    
        /*Do transfomation of the data to send on th enetwork and ensure the size of each varaibale
            is fixed whatever the system (32 bits or 64 bits or...)*/ 
        data_size_t=htonl(ADOC_global->header_window[ADOC_global->header_r].data_size);
        level_t=(uint8_t)(ADOC_global->header_window[ADOC_global->header_r].level);
        type_t=(uint8_t)(ADOC_global->header_window[ADOC_global->header_r].type);
        iov[0].iov_base=(void*)&(type_t);
        iov[0].iov_len=1;
        iov[1].iov_base=(void*)&(level_t);
        iov[1].iov_len=1;
        iov[2].iov_base=(void*)&(data_size_t);
        iov[2].iov_len=4;
        /*fprintf(stderr,"type:%d, level:%d\n",ADOC_global->header_window[ADOC_global->header_r].type,
                ADOC_global->header_window[ADOC_global->header_r].level);*/
                

        
#ifdef DUMP_PACKET

        printf("Header[%d]: type:%d level: %d size: %ld\n",ADOC_global->header_r,
               ADOC_global->header_window[ADOC_global->header_r].type,
               ADOC_global->header_window[ADOC_global->header_r].level,
               (long)data_size);
        if(iov[3].iov_len){
            fprintf(stderr,"**Data:");
            for(i=0;i<2;i++)
                fprintf(stderr,"%d (%ld) ",
                        (byte)(*((byte*)(&iov[3].iov_base[i]))),
                        (long)(&(iov[3].iov_base[i])));
            fprintf(stderr,"\n");
        }
#endif


#ifdef DEBUG_TIMEOUT
        {
            struct timeval tv;
            fd_set rfds;
            int res;
            int c=0;
            do{
                /* Pendant 1 secondes maxi */
                tv.tv_sec = 1;
                tv.tv_usec = 0;
                FD_ZERO(&rfds);
                FD_SET(socket, &rfds);
                res=select(socket+1,NULL,&rfds,NULL,&tv);
                if(!res){
                    c ++;
                    fprintf(stderr,"AdOC : select writev timeout : %d!\n",c);
                }

            }while(res<=0);
        }
#endif
        /*send information of the packet with data of the previous one*/

           
       if(start_clock){
            start_clock=0;
            nb_packets=0;
            CLOCK(start);
        }
        res=writev(socket,iov,4);
        

        if(res==-1){
            ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
            rval=adoc_sem_post(&(ADOC_global->history_sem));
            CHECK_RETVAL(rval,"sem_post");
            pthread_exit(0);
        }

        type=ADOC_global->header_window[ADOC_global->header_r].type;
 
        /*timings of buffer sending*/

        if(((prev_type==ADOC_TYPE_BS)||(prev_type==ADOC_TYPE_SESSION_SIZE))&&((type==ADOC_TYPE_UNCOMP)||(type==ADOC_TYPE_COMP)||(type==ADOC_TYPE_LAST_PACKET))){
            start_clock=1;
            buffer_size=0;
            sent_size=0;
            
        }

        sent_size+=prev_data_size;
        switch(prev_type){
            case ADOC_TYPE_LAST_PACKET:
            case ADOC_TYPE_COMP:
            case ADOC_TYPE_UNCOMP:
                nb_packets++;
        }
        
        level=ADOC_global->header_window[ADOC_global->header_r].level;
        
        if((prev_type==ADOC_TYPE_LAST_PACKET)||((type!=ADOC_TYPE_UNCOMP)&&(prev_type==ADOC_TYPE_UNCOMP))){
            if((prev_level!=level)||(buffer_size>=2*ADOC_BUFFER_SIZE)){
                CLOCK(end);
                start_clock=1;
                store_timing(ADOC_global,start,end,buffer_size,prev_level,nb_packets,sent_size);
                buffer_size=0;
                sent_size=0;
                            
            }
        }else if((prev_level==0)&&(buffer_size>=10*ADOC_PACKET_SIZE)){
            CLOCK(end);
            start_clock=1;
            store_timing(ADOC_global,start,end,buffer_size,prev_level,nb_packets,sent_size);
            buffer_size=0;
            sent_size=0;
        }

        if((type == ADOC_TYPE_UNCOMP)||(type==ADOC_TYPE_LAST_PACKET)){
            buffer_size+=ADOC_global->header_window[ADOC_global->header_r].uncomp_size;
        }
        



           prev_type=type;
        prev_level=ADOC_global->header_window[ADOC_global->header_r].level;
        switch(type){
            case ADOC_TYPE_LAST_PACKET:
            case ADOC_TYPE_COMP:
            case ADOC_TYPE_UNCOMP:
                /* data will be sent with information of the next paxket*/
                iov[3].iov_len=data_size;
                iov[3].iov_base=(void*)ADOC_global->header_window[ADOC_global->header_r].data;
                prev_data_size=data_size;
                ADOC_global->sent_size+=data_size;
#ifdef VERBOSE
                sent+=ADOC_global->header_window[ADOC_global->header_r].uncomp_size;
#endif
                break;
            case ADOC_TYPE_BS:
            case ADOC_TYPE_SESSION_SIZE: /*In that case data_size is the size of
                the whole data to be sent data*/
            case ADOC_TYPE_EOF:
                iov[3].iov_len=0;
                break;
            default :
                fprintf(stderr,"AdOC : Unknown packet type(%d)",type);
                ADOC_global->error=ADOC_UNKNOWN_PACKET;
                rval=adoc_sem_post(&(ADOC_global->history_sem));
                CHECK_RETVAL(rval,"sem_post");
                pthread_exit(0);
                /* Make compiler happy!*/
                return NULL;
        }
        /* Saying that one new packet can be added in the window*/
        rval=adoc_sem_post(&(ADOC_global->w_sem));
        CHECK_RETVAL(rval,"sem_post");
        pthread_mutex_lock (&(ADOC_global->fifo_mutex));
        ADOC_global->fifo_size--;
        pthread_mutex_unlock (&(ADOC_global->fifo_mutex));

        /*Increamentaing the reading pointer of the window cyclically*/
        ADOC_global->header_r++;
        if(ADOC_global->header_r>=ADOC_MAX_SEND_PACKETS)
            ADOC_global->header_r=0;
#ifdef VERBOSE
        printf("\r %.2f%% sent",((double)sent/ADOC_global->file_size)*100);
        fflush(stdout);
#endif
        /*Check whenever an error has been generated by
            the compression thread*/
        if(ADOC_global->error!=0){
            pthread_exit(0);
        }

    }while(type!=ADOC_TYPE_EOF);
    /* Telling the other thread that it can display the statistics*/
    rval=adoc_sem_post(&(ADOC_global->history_sem));
    CHECK_RETVAL(rval,"sem_post");
    pthread_exit(0);
    /* Make compiler happy!*/
    return NULL;
}


