#include "common.h"
#include "reception.h"
#include "timings.h"

ssize_t my_write(int fd,void * buf, size_t count,int line,char *filename){
    ssize_t nb;
    nb=write(fd,buf,count);
    if(nb<0){
        fprintf(stderr,"write error line %d of file %s : %s\n",
                line,filename,strerror(errno));
        fprintf(stderr,"count=%d, writen=%d\n",(int)count,(int)nb);
        perror("write");
    }

    if(nb!=count)
      fprintf(stderr,"########count=%d, writen=%d\n",(int)count,(int)nb);

    return nb;
}

ssize_t my_read(int fd, void* buffer, size_t count,int line,char *filename){
    ssize_t sum=0,nb=0;
    byte* buf=(byte*)buffer;
    do{
	jump : 
	nb=read(fd,buf+sum,count-sum); 
	if((nb==-1)&&(errno==EINTR))
	    goto jump;
	if(nb<0){
	    fprintf(stderr,"read error line %d of file %s: %s\n",
		    line,filename,strerror(errno));
	    return nb;
	}else if(nb==0) /*connection closed by peer*/
	  return sum;
    sum+=nb;
    }while(sum<count);

    return sum;
}



int adoc_close(int socket){
    free_buffer(socket);
    return close(socket);
}


int adoc_rec_init(ADOC_session *ADOC_global){

    /*shared objects initialization*/
    adoc_sem_init (&(ADOC_global->r_sem), 0, 0);
    /* Data arrive with the next packet, therefore we need one more packet to store the
    data*/
    adoc_sem_init (&(ADOC_global->w_sem), 0, ADOC_MAX_RECV_PACKETS-1);

    ADOC_global->header_window=(header*)malloc(sizeof(header)*(ADOC_MAX_RECV_PACKETS));
    if(ADOC_global->header_window==NULL){
	errno=ENOMEM;
	ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
	return -1;
    }
    ADOC_global->data_window=(byte *)malloc(ADOC_PACKET_SIZE*(ADOC_MAX_RECV_PACKETS)+2*ADOC_BUFFER_SIZE);
    if(ADOC_global->data_window==NULL){
	errno=ENOMEM;
	ADOC_global->error=ADOC_ERROR_WITH_ERRNO_SET;
	free(ADOC_global->header_window);
	return -1;
    }
    ADOC_global->header_r=0;
    ADOC_global->header_w=0;
    ADOC_global->data_w_pointer=ADOC_global->data_window;
    ADOC_global->ADOC_flag=0; 
    ADOC_global->error=0;
    return 0;
}


void adoc_rec_finalyze(ADOC_session *ADOC_global){
    free(ADOC_global->header_window);
    free(ADOC_global->data_window);
    adoc_sem_destroy((&ADOC_global->r_sem));
    adoc_sem_destroy(&(ADOC_global->w_sem));

}

ssize_t raw_recv_file(int socket,FILE* pf,size_t data_size){
    ssize_t count;
    size_t nb_recv=0;
    byte buffer[ADOC_PACKET_SIZE];
    while(data_size>nb_recv){
        count=READ(socket,buffer,ADOC_PACKET_SIZE);
        if(count<=0)
            return nb_recv;
        write(fileno(pf),buffer,count);
        nb_recv+=count;
    }
    return nb_recv;
}

ssize_t  adoc_receive_file(int socket,FILE *pf){
    pthread_t th1,th2;
    void *ret;
    ssize_t res=0;
    ADOC_session ADOC_global;
    uint32_t sval;
    size_t val;
    int8_t c;

    /*can we do raw recv? */
    res=READ(socket,(byte *)&c,1);
    if(res==0)
        return 0;
    if(c>=0)
        return raw_recv_file(socket,pf,c);

    switch (c){
        case ADOC_RAW:
            READ(socket,(byte *)&sval,4);
            val=(size_t)ntohl(sval);
            return raw_recv_file(socket,pf,val);
        case ADOC_CMP:
            break;
        case ADOC_TEST:
            
            READ(socket,(byte *)&c,1);

            if(c>=0)
                val=c;
            else{
                READ(socket,(byte *)&sval,4);
                val=(size_t)ntohl(sval);
            }
            res=raw_recv_file(socket,pf,val);
            if(res<0)
                return res;
            READ(socket,(byte *)&c,1);
            if(c>=0)
                return raw_recv_file(socket,pf,c)+res;
            if(c==ADOC_RAW){
                READ(socket,(byte *)&sval,4);
                val=(size_t)ntohl(sval);
                return raw_recv_file(socket,pf,val)+res;
            }else if(c==ADOC_CMP)
                break;
            else
                return ADOC_PROTOCOL_ERROR;
        default:
            return ADOC_PROTOCOL_ERROR;
    }
    
    if(adoc_rec_init(&ADOC_global)!=0)
	return ADOC_global.error;
    ADOC_global.file=pf;
    ADOC_global.socket=socket;
    ADOC_global.byte_recv=&res;
    if (pthread_create (&th1, NULL, file_decompression_thread, &ADOC_global) < 0) {
	fprintf (stderr, "pthread_create error for decompression thread\n");
	free(ADOC_global.header_window);
	free(ADOC_global.data_window);	
	return -1;
    }
    
    if (pthread_create (&th2, NULL, reception_thread, &ADOC_global) < 0) {
	fprintf (stderr, "pthread_create error for reception thread\n");
	free(ADOC_global.header_window);
	free(ADOC_global.data_window);	
	return -1;
    }

    (void)pthread_join (th1, &ret);
    (void)pthread_join (th2, &ret);
    adoc_rec_finalyze(&ADOC_global);
    if(ADOC_global.error==0)
	return res;
    else 
	return ADOC_global.error;
}


ssize_t launch_recep_threads(int socket,byte *data_buffer, size_t count,byte *s_buffer, size_t store_size,ssize_t res,size_t size){
    
    pthread_t th1,th2;
    void *p_ret;
    ssize_t recv_size;
    
    ADOC_session ADOC_global;
    ADOC_global.store_buffer=s_buffer;
    ADOC_global.store_size=store_size;

    if(adoc_rec_init(&ADOC_global)!=0)
        return ADOC_global.error;

    recv_size=res;
    ADOC_global.data_buffer=data_buffer;
    ADOC_global.socket=socket;
    ADOC_global.count=count;
    ADOC_global.byte_recv=&recv_size;	

    if ((res=pthread_create (&th1, NULL, data_decompression_thread, &ADOC_global)) < 0) {
        fprintf (stderr, "pthread_create error for decompression thread %ld(%d) :%s\n",(long)res,errno,strerror(errno));
        free(ADOC_global.header_window);
        free(ADOC_global.data_window);
        return -1;
    }
    if ((res=pthread_create (&th2, NULL, reception_thread, &ADOC_global)) < 0) {
        fprintf (stderr, "pthread_create error for reception thread : %s\n",strerror(errno));
        free(ADOC_global.header_window);
        free(ADOC_global.data_window);	
        return -1;
    }
    (void)pthread_join (th1, &p_ret);
    (void)pthread_join (th2, &p_ret);
    adoc_rec_finalyze(&ADOC_global);
    if(ADOC_global.error==0)
        return recv_size+size;
    else 
        return ADOC_global.error;
}




ssize_t adoc_read(int socket,void *buffer,
                  size_t count){

    ssize_t res;

    res=adoc_read_wraper(socket,buffer,count);

    /* Debug timing
      {
	CLOCK_T date;
	CLOCK(date);
	CLOCK_DISPLAY(date); fprintf(stderr," : read : %ld\n",(long int)res);
	}
    */
    return res;
}






