
#include <stdlib.h>
#include "adoc.h"
#include "timings.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <netinet/tcp.h>
#include <inttypes.h>
#include "client-server.h"


#undef LATENCY_ONLY
#define CHECK_ADOC_RETVAL(a,b) check_adoc_retval(a,b,__LINE__);

typedef int8_t byte;
typedef uint8_t ubyte;

#define NAME_SIZE  1024

void check_adoc_retval(ssize_t ret,size_t count,int line){
    if(ret<0){
	fprintf(stderr,"AdOC Error line %d",line);
	switch(ret){
	case -1: 
	    perror("");
	    break;
	case ADOC_UNKNOWN_PACKET :
	    fprintf(stderr,": Unknown packet ID\n");
	    break;
	case ADOC_COMPRESSION_ERROR :
	    fprintf(stderr,": error while compressing or uncompressing data\n");
	    break;
	case ADOC_PROTOCOL_ERROR :
	    fprintf(stderr,": error in the ADOC protocol\n");
	    break;
	case ADOC_LEVEL_ERROR :
	    fprintf(stderr,": unexpected compression level\n");
	    break;
	case ADOC_NETWORK_ERROR :
	    fprintf(stderr,": wrong size of data received\n");
	    break;
	case ADOC_WRITE_ERROR :
	    fprintf(stderr,": wrong size of data writen by system call write\n");
	    break;
	case ADOC_RETVAL_ERROR :
	    fprintf(stderr,": return Value != 0\n");
	    break;
	case ADOC_FILE_ERROR :
	    fprintf(stderr,": error while reading file\n");
	    break;
	default :
	    fprintf(stderr,": unknown error\n");
	    break;
	}
	exit(-1);
    }

    if(ret!=count){
	fprintf(stderr,"AdOC Error line %d: Number of bytes sent or received different from count value\n",line);
	exit(-1);
    }
}


/*Create a socket TCP and return port number*/
int create_sock(int *port, int type){

  int desc; /*dsocket descriptor*/
  struct sockaddr_in nom; /*socket adress*/
  socklen_t longueur; /* adress length*/

  /* creating socket */

  if((desc=socket(AF_INET,type,0))==-1){
    perror("Socket creation failed.\n");
    return -2;
  }
  

  nom.sin_port=htons(*port);
  nom.sin_addr.s_addr=INADDR_ANY;
  nom.sin_family=AF_INET;

  if(bind(desc,(struct sockaddr *)&nom,sizeof(nom))){
    perror("Socket naming failed.\n");
    return -3;
  }
  
  longueur = sizeof(nom);
  if(getsockname(desc,(struct sockaddr *)&nom,&longueur)){
    perror("Getting name of the socket failed.\n");
    return -4;
  }

  *port=ntohs(nom.sin_port);


  return desc;
}


int time_transmission(int socket,ubyte * buffer){

    ssize_t cc;
    CLOCK_T begin,end;
    double time;
    int i;

    bzero(buffer,BUFFER_SIZE);

    printf("Reading buffer...");fflush(stdout);
    
    CLOCK(begin);

    cc=adoc_read(socket,buffer,BUFFER_SIZE);
    CLOCK(end);  
    CHECK_ADOC_RETVAL(cc,BUFFER_SIZE);

    printf("OK!\n");
    
    time=CLOCK_DIFF(end,begin);
    
    printf("Receiving %ld bytes in time =%.2f seconds\n",(long)cc,time);
    printf("Checking data...");
    for(i=0;i<BUFFER_SIZE;i++)
	if(buffer[i]!=i%10){
	printf("\nBuffer[%d]=%d error\n",i,buffer[i]);
	return 0;
	}
    printf("OK!\n No error.\n");



    bzero(buffer,BUFFER_SIZE);
    printf("Reading buffer...");fflush(stdout);
    CLOCK(begin);
    cc=adoc_read(socket,buffer,BUFFER_SIZE);
    CLOCK(end);  
    CHECK_ADOC_RETVAL(cc,BUFFER_SIZE);

    printf("OK!\n");
    time=CLOCK_DIFF(end,begin);
    printf("Receiving %ld bytes in time =%.2f secondes\n",(long)cc,time);
    printf("Checking data...");
    for(i=0;i<BUFFER_SIZE;i++)
	if(buffer[i]!=i%10){
	printf("\nBuffer[%d]=%d error\n",i,buffer[i]);
	return 0;
	}
    printf("OK!\n No error.\n");

    bzero(buffer,BUFFER_SIZE);
    printf("Reading buffer...");fflush(stdout);
    CLOCK(begin);
    cc=adoc_read(socket,buffer,BUFFER_SIZE);
    CLOCK(end);  
    CHECK_ADOC_RETVAL(cc,BUFFER_SIZE);

    printf("OK!\n");
    time=CLOCK_DIFF(end,begin);
    printf("Receiving %ld bytes in time =%.2f secondes\n",(long)cc,time);
    printf("Checking data...");
    for(i=0;i<BUFFER_SIZE;i++)
	if(buffer[i]!=i%10){
	printf("\nBuffer[%d]=%d error\n",i,buffer[i]);
	return 0;
	}
    printf("OK!\n No error.\n");
    return 1;
}

void time_latency(int socket,ubyte * buffer){
    size_t sent_size;
    int i;

    printf("Timing Latency...\n");

    for(i=0;i<50;i++){
        adoc_read(socket,buffer,1);
        adoc_write(socket,buffer,1,&sent_size);    
    }

    for(i=0;i<50;i++){
        adoc_read(socket,buffer,1);
        adoc_write_levels(socket,buffer,1,&sent_size,1,6);
    }

    for(i=0;i<50;i++){
        read(socket,buffer,1 );
        write(socket,buffer,1);
    }
}

int ping_pong_test(int socket,ubyte *buffer){
    ubyte tab[100];
    int err_flag=0,j,i;
    size_t data_size,sent_size;
    ssize_t cc;

    cc=adoc_read(socket,tab,100);
    if(cc!=100)
	fprintf(stderr,"Receive error cc=%ld\n",(long)cc);

    for(i=0;i<100;i++){
	data_size=(size_t)tab[i];
	if((data_size%2)){
	    printf("s(%lu)",(unsigned long)data_size);fflush(stdout);
	    memset(buffer,1,data_size);
	    cc=adoc_write(socket,buffer,data_size,&sent_size);
	    printf(". ");fflush(stdout);
            if(cc!=data_size){
		fprintf(stderr,"\nError for step %d\n",i);
		err_flag++;
	    }
	    
	    for(j=0;j<data_size;j++)
		if(buffer[j]!=1){
		    fprintf(stderr,"\nbuffer[%d]=%d corrupted!\n",j,buffer[j]);
		    err_flag++;
		}
	}else{
	    printf("r(%lu)",(unsigned long)data_size);fflush(stdout);
	    bzero(buffer,data_size);
	    cc=adoc_read(socket,buffer,data_size);
	    printf(". ");fflush(stdout);

	    if(cc!=data_size){
		err_flag++;
		fprintf(stderr,"\nError for step %d\n",i);
	    }
	    
	    for(j=0;j<data_size;j++)
		if(buffer[j]!=1){
		    fprintf(stderr,"\nbuffer[%d]=%d error!\n",j,buffer[j]);
		    err_flag++;
		}
	}
    }
    return err_flag;
}

/* with this test partial read are performed.
   AdOC must store in temporary buffers extra data that are received but not read
   These data must be retreived at the next call before starting to read the socket
   This test check that all goes right with these temporary buffers
   when no compression is performed*/
int large_buffer_no_cmp_test(int socket,ubyte *buffer){
    int err_flag=0,i;
    ssize_t cc,cc1,cc2;

    printf("-------Test of internal buffers without compression----------\n");
    err_flag=0;
    bzero(buffer,100000);
    cc=adoc_read(socket,buffer,400);
    cc1=adoc_read(socket,buffer+cc,50000);
    cc2=adoc_read(socket,buffer+cc1+cc,100000-cc1-cc);

    if(cc1+cc2+cc!=100000){
	err_flag++;
        fprintf(stderr,"\nError cc=%d shoud be 4000 ",(int)cc);
        fprintf(stderr,"or cc1=%d shoud be 50000",(int)cc1);
        fprintf(stderr,"or cc2=%d shoud be 46000",(int)cc2);
    }

    for(i=0;i<100000;i++)
	if(buffer[i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i,buffer[i]);
            break;
	}

    
   if(err_flag==0)
        printf("\nOK\n No error while reading 1 sent buffer with many read.\n");
   else{
	printf("\nNo ok! %d error!\nInternal buffer tests without comression failed",err_flag);
	return err_flag;
   }
       bzero(buffer,200000);
   cc=adoc_read(socket,buffer,10000);
   cc1=adoc_read(socket,buffer+cc,10000);
   cc2=adoc_read(socket,buffer+cc+cc1,200000-cc-cc1);
    for(i=0;i<100000;i++)
	if(buffer[i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i,buffer[i]);
	    return err_flag;
	}

   if(err_flag==0)
        printf("\nOK!\nNo error while partially read a sent buffer.\n");
   else{
	printf("\nNo ok! %d error while partially read a sent buffer!\n",err_flag);
	return err_flag;
   }

    for(i=0;i<100000;i++)
	if(buffer[100000+i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i+10000,buffer[i+100000]);
	    return err_flag;
	}
   if(err_flag==0)
        printf("\nOK!\nNo error while using internal buffers and incomming data for reading incomming data.\n");
   else{
	printf("\nNo ok! %d error!\n",err_flag);
	return err_flag;

   }
   printf("\nread and write unix system call semantic test passed with desabled compression.\n");
   return err_flag;
}


/* with this test partial read are performed.
   AdOC must store in temporary buffers extra data that are received but not read
   These data must be retreived at the next call before starting to read the socket
   This test check that all goes right with these temporary buffers 
   when compression is performed*/
int large_buffer_cmp_test(int socket,ubyte *buffer){
    int err_flag=0,i;
    ssize_t cc,cc1,cc2;
	
   printf("\n\n-------Test of internal buffers with compression----------\n");
		
    err_flag=0;
    bzero(buffer,900000);
    cc=adoc_read(socket,buffer,4000);
    cc1=adoc_read(socket,buffer+cc,50000);
    cc2=adoc_read(socket,buffer+cc1+cc,900000-cc1-cc);

    if(cc1+cc2+cc!=900000){
	err_flag++;
        fprintf(stderr,"\nError cc=%d shoud be 4000 ",(int)cc);
        fprintf(stderr,"or cc1=%d shoud be 50000",(int)cc1);
        fprintf(stderr,"or cc2=%d shoud be 46000",(int)cc2);
    }

    for(i=0;i<900000;i++)
	if(buffer[i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i,buffer[i]);
	    fprintf(stderr,"\nbuffer[%d]=%d \n",i-1,buffer[i-1]);
	    return err_flag;
	}

    
   if(err_flag==0)
        printf("\nOK!\n No error while reading 1 sent buffer with many read.\n");
   else{
	printf("\nNo ok! %d error!\nInternal buffer tests with disabled compression failed",err_flag);
	return err_flag;
   }
   
   bzero(buffer,1800000);
    cc=adoc_read(socket,buffer,100000);
    cc=adoc_read(socket,buffer+100000,1700000);
    for(i=0;i<900000;i++)
	if(buffer[i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i,buffer[i]);
	    return err_flag;
	}

   if(err_flag==0)
        printf("\nOK!\nNo error while partially read a sent buffer.\n");
   else{
	printf("\nNo ok! %d error while partially read a sent buffer!\n",err_flag);
	return err_flag;
   }
    for(i=0;i<900000;i++)
	if(buffer[900000+i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i+900000,buffer[i+900000]);
	    return err_flag;
	}
   if(err_flag==0)
        printf("\nOK!\nNo error while using internal buffers and incomming data for reading incomming data.\n");
   else{
	printf("\nNo ok! %d error!\n",err_flag);
	return err_flag;

   }
   printf("\nread and write unix system call semantic test passed with forced compression.\n");
   return err_flag;
}




/* with this test partial read are performed.
   AdOC must store in temporary buffers extra data that are received but not read
   These data must be retreived at the next call before starting to read the socket
   This test check that all goes right with these temporary buffers 
   for small buffers*/
int small_buffer_test(int socket,ubyte *buffer){
    int err_flag=0;
    ssize_t cc,cc1,cc2;
    int i;
 printf("-------Test for small read buffers----------\n");
    err_flag=0;
    bzero(buffer,100000);
    cc=adoc_read(socket,buffer,5);
    cc1=adoc_read(socket,buffer+cc,4);
    cc2=adoc_read(socket,buffer+cc1+cc,100000-cc1-cc);

        
    if(cc1+cc2+cc!=100000){
	err_flag++;
        fprintf(stderr,"\nError cc=%d shoud be 5 ",(int)cc);
        fprintf(stderr,"or cc1=%d shoud be 4",(int)cc1);
        fprintf(stderr,"or cc2=%d shoud be 99991",(int)cc2);
    }

    for(i=0;i<100000;i++)
	if(buffer[i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i,buffer[i]);
            break;
	}

    
   if(err_flag==0)
        printf("\nOK\n No error while reading 1 sent buffer with many small read.\n");
   else{
	printf("\nNo ok! %d error!\nInternal buffer tests without comression failed",err_flag);
       	return err_flag;
   }
        printf("-------Test for small read buffers and forced compression----------\n");
    err_flag=0;
    bzero(buffer,100000);
    cc=adoc_read(socket,buffer,5);
    cc1=adoc_read(socket,buffer+cc,4);
    cc2=adoc_read(socket,buffer+cc1+cc,100000-cc1-cc);

    if(cc1+cc2+cc!=100000){
	err_flag++;
        fprintf(stderr,"\nError cc=%d shoud be 5 ",(int)cc);
        fprintf(stderr,"or cc1=%d shoud be 4",(int)cc1);
        fprintf(stderr,"or cc2=%d shoud be 99991",(int)cc2);
    }

    for(i=0;i<100000;i++)
	if(buffer[i]!=i%250){
	    err_flag++;
	    fprintf(stderr,"\nbuffer[%d]=%d error!\n",i,buffer[i]);
            break;
	}

    
   if(err_flag==0)
        printf("\nOK\n No error while reading 1 sent buffer with many small read.\n");
   else{
	printf("\nNo ok! %d error!\nInternal buffer tests without comression failed",err_flag);
	return err_flag;
   }
   return err_flag;
}

int closed_socket_test(int socket,ubyte *buffer){
    ssize_t cc;

   printf("\nLast check... should not be longer than 1 seconds\n");
   
   /*wait for the client to close the connexion*/


   printf("Reading data from a socket closed by the peer...\n");
   printf("If the program does not output \"All tests passed!\" immediatly : there is a bug!\n");
   cc=adoc_read(socket,buffer,10);
   if(cc!=0){
       printf("Wrong returned value : %ld\n",(long)cc);
       return 0;
   }
   return 1;
}

/* Service for receiving data*/
void service(int socket){
 
    ubyte *buffer;
    int res,err_flag;
    
    buffer=(ubyte *)malloc(BUFFER_SIZE*sizeof(ubyte));
 
    res=time_transmission(socket,buffer);
    if(!res){
	free(buffer);
	exit(-1);
    }
    
    time_latency(socket,buffer);
    

    err_flag=ping_pong_test(socket,buffer);

    if(err_flag==0)
        printf("\nOK!\nping-pong tests passed.\n");
    else{
	printf("\nNo ok! %d error!\nPing pong tests failed.",err_flag);
	free(buffer);
	return;
    }


    res=large_buffer_no_cmp_test(socket,buffer);

    if(res!=0){
	free(buffer);
	exit(-1);
    }


    res=large_buffer_cmp_test(socket,buffer);

    if(res!=0){
	free(buffer);
	exit(-1);
    }

    res=small_buffer_test(socket,buffer);

    if(res!=0){
	free(buffer);
	exit(-1);
    }

    
    res=closed_socket_test(socket,buffer);
    if(!res){
	free(buffer);
	exit(-1);
    
    }

   printf ("\nAll tests passed!\n");
   printf("Do ctrl-C if you whish to terminate the server, start an other client otherwise\n");
   free(buffer);
}



void wait_sigchild(int sig){
    int n,code;
    n=wait(&code);
    printf("My child number %d has sent signal %d and terminated with exit code %d!\n",n,sig,code>>8); 
}


int main(int argc, char **argv){
    
  int listen_sock,service_sock;
  struct sockaddr_in adr;
  socklen_t lgadr = sizeof(struct sockaddr_in);
  char name[NAME_SIZE];
  int port=0;

   if(argc==2){
      port=atoi(argv[1]);
      printf("Trying port : %d\n",port);
  }


  if((listen_sock=create_sock(&port,SOCK_STREAM))==-1){
    fprintf(stderr,"Creation/liason de socket impossible\n");
    exit(2);
  }

  gethostname(name,NAME_SIZE);

  printf("Socket open on port %d on host %s\n",port,name);


  /*creation of the pending connection file */

  listen (listen_sock,5);

  /*Handler for SIGCHLD signal
    signal (SIGCHLD,wait_sigchild);*/

  while(1){

    lgadr=sizeof(adr);
    /* waiting for a connection*/
    if((service_sock=accept(listen_sock,(struct sockaddr*)&adr,&lgadr))<0){
	fprintf(stderr,"accept error :%s\n",strerror(errno));
	continue;
    }
    
    /* forking processus*/
    if(fork()==0){
	/* son executes service*/
      close(listen_sock);
      service(service_sock);
      exit(0);
    } else{
	/*father return to accept*/
	close(service_sock);
    }
    
  }
  return 0;
}












